# -*- coding: utf-8 -*-

"""Context manager for extending file reading to handle include's"""

import bz2
import collections
import gzip
import logging
import os
import os.path

module_logger = logging.getLogger(__name__)


def splitext(filename):
    """
    Get the extension of a file, ignoring .gz or .bz2 on the end

    Args:
        filename ('str'): the name, including any path, of the file
    """
    name, ext = os.path.splitext(filename)
    if ext in ('.gz', '.bz2'):
        name, ext = os.path.splitext(name)
    return ext


class Open(object):

    def __init__(
        self, filename, mode='r', logger=None, include=None, history=10
    ):
        """Open a file, automatically handling 'include'

        Args:
            filename ('str'): The path to the file
            mode ('str', optional): The mode to open the file. Defaults to
                read-only.
            logger (obj, optional): The logging object to use
            include ('str', optional): The keyword that triggers an include
            history (integer, optional): Length of history to keep
        """
        if not isinstance(filename, str):
            raise RuntimeError("Filename must be a string path")
        self._filename = filename
        self.mode = mode
        if logger:
            self.logger = logger
        else:
            self.logger = module_logger
        self.include = include
        self._history = history
        self._depth = -1

        self._deque = collections.deque(maxlen=history)

        self._total_lines = 0
        self._linenos = []
        self._files = []
        self._paths = []
        self._fds = []
        self._cwd = None

    def __enter__(self):
        """Handle the enter event for the context manager by opening the file
        """
        self.logger.debug('in __enter__')
        self._cwd = os.getcwd()
        self._linenos.append(0)
        self._files.append(self._filename)
        filepath = os.path.join(self._cwd, self._filename)
        path = os.path.dirname(filepath)
        self._paths.append(path)

        self._fds.append(self._open(filepath))

        self.logger.debug('   opened {}'.format(self._filename))
        return self

    def __exit__(self, *args, **kwargs):
        """Close any open files as we exit the context
        """
        self.logger.debug('in __exit__')
        while len(self._fds) > 0:
            fd = self._fds.pop()
            self._close(fd, *args, **kwargs)
        self.logger.debug('   closed all files')

    def __next__(self):
        """Iterator to get the next line
        """
        self.logger.log(0, "__next__")

        if self._depth >= 0:
            line = self._deque[self._depth]
            self._depth -= 1
            return line

        line = self._next()
        words = line.split()
        if self.include and len(words) > 0 and words[0] == self.include:
            filename = line.split()[1]
            self.logger.debug('   opening include file {}'.format(filename))
            self._linenos.append(0)
            self._files.append(filename)
            filepath = os.path.join(self._paths[-1], filename)
            path = os.path.dirname(filepath)
            self._paths.append(path)

            self._fds.append(self._open(filepath))

            self.logger.debug('   opened it')
            line = self.__next__()
        self.logger.log(0, line)

        self._deque.appendleft(line)
        return line

    def __iter__(self):
        """Need to be an iterator
        """
        self.logger.log(0, "__iter__")
        return self

    def __getattr__(self, attr):
        """Pass any attribute requests to the actual file handle
        """
        self.logger.debug("attr = '{}'".format(attr))
        return getattr(self._fds[-1], attr)

    @property
    def path(self):
        if len(self._paths) > 0:
            return self._paths[-1]
        else:
            return None

    @property
    def filename(self):
        if len(self._files) > 0:
            return self._files[-1]
        else:
            return None

    @property
    def lineno(self):
        if len(self._linenos) > 0:
            return self._linenos[-1]
        else:
            return 0

    @property
    def total_lines(self):
        return self._total_lines

    @property
    def depth(self):
        return self._depth

    def push(self, n=1):
        """
        push the n last lines back onto the device (virtually)
        """
        if self._depth + n > len(self._deque):
            raise RuntimeError("Exceeded the currently available history")
        self._depth += n

    def stack(self):
        """Provide the traceback of the included files
        """
        result = []
        for i in range(len(self._paths) - 1, 0, -1):
            path = self._paths[i]
            filename = self._files[i]
            filepath = os.path.join(path, filename)
            lineno = self._linenos[i]
            result.append('{}:{}'.format(filepath, lineno))
        return result

    def _close(self, fd, *args, **kwargs):
        """Helper routine to close a file handle
        """
        exit = getattr(fd, '__exit__', None)
        if exit is not None:
            self.logger.debug('   closing fd using its __exit__ method')
            return exit(*args, **kwargs)
        else:
            exit = getattr(fd, 'close', None)
            if exit is not None:
                self.logger.debug('   closing fd using its close method')
                exit()

    def _next(self):
        """Helper routine to get the next line, handling EOF and errors
        """
        try:
            line = self._fds[-1].__next__()
        except StopIteration:
            fd = self._fds.pop()
            self._close(fd)
            self._files.pop()
            self._paths.pop()
            n = self._linenos.pop()
            self.logger.info('   read {} lines from fd'.format(n))
            if len(self._fds) <= 0:
                raise StopIteration()
            return self._next()
        except:  # noqa: E722
            raise
        self._linenos[-1] += 1
        self._total_lines += 1
        return line

    def _open(self, filename):
        """Open 'filename' using gzip or bzip if needed

        Args:
            filename ('str'): the name, including any path, of the file
        """
        if not isinstance(filename, str):
            raise RuntimeError("Filename must be a string to open")
        name, ext = os.path.splitext(filename.strip().lower())
        if ext == '.bz2':
            fd = bz2.open(filename, self.mode + 't')
        elif ext == '.gz':
            fd = gzip.open(filename, self.mode + 't')
        else:
            fd = open(filename, self.mode)
        return fd



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package, units module."""

from seamm_util import Q_


def test_substance():
    """Testing that we can convert amounts of substance."""
    E = Q_(1.0, 'mol')
    E2 = E.to('')
    assert f'{E:~} = {E2:~.4}' == '1.0 mol = 6.022e+23'


def test_inverse_substance():
    """Testing that we can convert amounts of substance."""
    E = Q_(6.022e+23, '')
    E2 = E.to('mol')
    assert f'{E:~} = {E2:~.4}' == '6.022e+23 = 1.0 mol'


def test_energy():
    """Testing that we can convert amounts of substance."""
    E = Q_(1.0, 'kcal/mol')
    E2 = E.to('eV')
    assert f'{E:~} = {E2:~.4}' == '1.0 kcal / mol = 0.04336 eV'


def test_inverse_energy():
    """Testing that we can convert amounts of substance."""
    E = Q_(1.0, 'eV')
    E2 = E.to('kcal/mol')
    assert f'{E:~} = {E2:~.4}' == '1.0 eV = 23.06 kcal / mol'


def test_force():
    """Testing that we can convert amounts of substance."""
    E = Q_(1, 'kcal/mol/angstrom')
    E2 = E.to('eV/angstrom')
    assert f'{E:~} = {E2:~.4}' == '1 kcal / mol / Å = 0.04336 eV / Å'


def test_inverse_force():
    """Testing that we can convert amounts of substance."""
    E = Q_(1, 'eV/angstrom')
    E2 = E.to('kcal/mol/angstrom')
    assert f'{E:~} = {E2:~.4}' == '1 eV / Å = 23.06 kcal / mol / Å'



# -*- coding: utf-8 -*-

"""Localize the unit handling."""

import pint

# Unit handling!
ureg = pint.UnitRegistry(auto_reduce_dimensions=True)
pint.set_application_registry(ureg)
Q_ = ureg.Quantity
units_class = ureg('1 km').__class__

_d = pint.Context('chemistry')

factor = ureg.mol / ureg.avogadro_number

_d.add_transformation('', '[substance]', lambda ureg, x: x * factor)
_d.add_transformation('[substance]', '', lambda ureg, x: x / factor)

_d.add_transformation(
    '[mass]/[substance]', '[mass]', lambda ureg, x: x * factor
)

_d.add_transformation(
    '[length] * [mass] / [substance] / [time] ** 2',
    '[length] * [mass] / [time] ** 2',
    lambda ureg, x: x * factor
)
_d.add_transformation(
    '[length] * [mass] / [time] ** 2',
    '[length] * [mass] / [substance] / [time] ** 2',
    lambda ureg, x: x / factor
)

_d.add_transformation(
    '[length] ** 2 * [mass] / [substance] / [time] ** 2',
    '[length] ** 2 * [mass] / [time] ** 2',
    lambda ureg, x: x * factor
)
_d.add_transformation(
    '[length] ** 2 * [mass] / [time] ** 2',
    '[length] ** 2 * [mass] / [substance] / [time] ** 2',
    lambda ureg, x: x / factor
)

ureg.add_context(_d)
ureg.enable_contexts('chemistry')


if __name__ == '__main__':  # pragma: no cover
    E = Q_(1.0, 'kcal/mol')
    E2 = E.to('eV')
    print(f'{E:~} = {E2:~.4}')
    print()

    E2 = E.to('kJ/mol')
    print(f'{E:~} = {E2:~.4}')
    print()

    E = Q_(1.0, 'eV')
    E2 = E.to('kcal/mol')
    print(f'{E:~} = {E2:~.4}')
    print()

    E2 = E.to('kJ/mol')
    print(f'{E:~} = {E2:~.4}')
    print()

    E = Q_(1.0, 'mol')
    E2 = E.to('')
    print(f'{E:~} = {E2:~.4}')
    print()

    E = Q_(6.022e+23, '')
    E2 = E.to('mol')
    print(f'{E:~} = {E2:~.4}')
    print()

    E = Q_(1, 'kcal/mol/angstrom')
    E2 = E.to('eV/angstrom')
    print(f'{E:~} = {E2:~.4}')
    print()

    E = Q_(1, 'eV/angstrom')
    E2 = E.to('kcal/mol/angstrom')
    print(f'{E:~} = {E2:~.4}')
    print()



# -*- coding: utf-8 -*-

"""A class for wrapping output from the MolSSI Framework and plugins.
Initially it is a thin wrapper around a dictionary to store the output,
using JSON to serialize and deserialize the data.

It presents the dict api with the addition of methods to serialize and
deserialize the contents.
"""

import bz2
import collections.abc
import gzip
import json
import os.path
import pprint


class Output(collections.abc.MutableMapping):

    def __init__(self, filename=None, **kwargs):
        """Create the Output object

        Keyword arguments:
            filename: the name of the file for serialization
            kwargs: any other keyword arguments initialize the dict
        """

        self._is_changed = False
        self._data = dict()  # This stores the dict like data
        self._filename = None

        # Set the filename, if given, which will open the file
        self.filename = filename

        # Finally, process any extra keywords
        if len(kwargs) > 0:
            self.update(dict(**kwargs))

    @property
    def filename(self):
        """The filename for the serialized data"""
        return self._filename

    @filename.setter
    def filename(self, filename):
        if filename != self._filename:
            if self._filename is None:
                if self._is_changed:
                    pass
            else:
                if self._is_changed:
                    self.flush()
            self._filename = filename

        if self._filename is not None:
            with self._open() as fd:
                self._data = json.load(fd)
            self._is_changed = False

    def __getitem__(self, key):
        """Allow [] access to the dictionary!"""
        if key not in self._data:
            raise KeyError("key '" + key + "' does not exist")
        return self.get(key)

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._data[key] = value
        self._is_changed = True

    def __delitem__(self, key):
        """Allow deletion of keys"""
        del self._data[key]
        self._is_changed = True

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._data)

    def __len__(self):
        """The len() command"""
        return len(self._data)

    def __repr__(self):
        """The string representation of this object"""
        return repr(self._data)

    def __str__(self):
        """The pretty string representation of this object"""
        return pprint.pformat(self._data)

    def copy(self):
        """Return a shallow copy of the dictionary"""
        return self._data.copy()

    def _open(self):
        """Open self.filename, using compression according to its extension,
        """

        extension = os.path.splitext(self.filename)[1].strip().lower()
        if extension == '.bz2':
            fd = bz2.open(self.filename, 'rt')
        elif extension == '.gz':
            fd = gzip.open(self.filename, 'rt')
        else:
            fd = open(self.filename, 'r')

        return fd



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package, plotting module."""

from seamm_util import Figure
import jinja2
import os
# import pytest

# Create the Jinja environment with the path we want
results_path = os.path.join(os.path.dirname(__file__), 'results')
template_path = os.path.join(os.path.dirname(__file__), 'templates')
env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_path))


def test_simple_plot():
    """Testing that we can create a simple plot."""
    fig = Figure(
        jinja_env=env, template='line.html_template', title='Simple plot'
    )

    plot = fig.add_plot('Energies')

    x_axis = plot.add_axis('x', label='Time (ps)')
    y_axis = plot.add_axis('y', label='Energy (kcal/mol)', anchor=x_axis)
    x_axis.anchor = y_axis

    plot.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name='Total',
        showlegend='false',
        x0=0,
        dx=1.0,
        xlabel='t',
        xunits='ps',
        y=[110.0, 111.0, 110.0, 109.0, 110.0],
        ylabel='E',
        yunits='kcal/mol',
        color='#20a8d8'
    )

    graph = fig.dumps()

    results_file = os.path.join(results_path, 'simple_plot.html')
    try:
        with open(results_file, 'r') as fd:
            correct_result = fd.read()
    except Exception:
        correct_result = ''

    if graph != correct_result:
        with open(results_file + '+', 'w') as fd:
            fd.write(graph)

    assert graph == correct_result


def test_two_plots():
    """Testing that we can create two plots."""
    fig = Figure(
        jinja_env=env, template='line.html_template', title='Two plots'
    )

    plot1 = fig.add_plot('plot1')

    x_axis = plot1.add_axis('x', label='Time (ps)')
    y_axis = plot1.add_axis('y', label='Energy (kcal/mol)', anchor=x_axis)
    x_axis.anchor = y_axis

    plot1.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name='Total',
        showlegend='false',
        x0=0,
        dx=1.0,
        xlabel='t',
        xunits='ps',
        y=[110.0, 111.0, 110.0, 109.0, 110.0],
        ylabel='E',
        yunits='kcal/mol',
        color='#20a8d8'
    )

    plot2 = fig.add_plot('plot2')

    x_axis = plot2.add_axis('x', label='Time (ps)')
    y_axis = plot2.add_axis('y', label='Energy (kcal/mol)', anchor=x_axis)
    x_axis.anchor = y_axis

    plot1.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name='Small',
        showlegend='false',
        x0=0,
        dx=1.0,
        xlabel='t',
        xunits='ps',
        y=[10.0, 11.0, 10.0, 9.0, 10.0],
        ylabel='E',
        yunits='kcal/mol',
        color='blue'
    )

    fig.grid_plots(
        'plot1   -     x',
        '  ^     ^   plot2'
    )  # yapf: disable

    graph = fig.dumps()

    results_file = os.path.join(results_path, 'two_plots.html')
    try:
        with open(results_file, 'r') as fd:
            correct_result = fd.read()
    except Exception:
        correct_result = ''

    if graph != correct_result:
        with open(results_file + '+', 'w') as fd:
            fd.write(graph)

    assert graph == correct_result


def test_writing_plot(tmpdir):
    """Testing that we can create write a plot to disk."""
    fig = Figure(
        jinja_env=env, template='line.html_template', title='Simple plot'
    )

    plot = fig.add_plot('Energies')

    x_axis = plot.add_axis('x', label='Time (ps)')
    y_axis = plot.add_axis('y', label='Energy (kcal/mol)', anchor=x_axis)
    x_axis.anchor = y_axis

    plot.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name='Total',
        showlegend='false',
        x0=0,
        dx=1.0,
        xlabel='t',
        xunits='ps',
        y=[110.0, 111.0, 110.0, 109.0, 110.0],
        ylabel='E',
        yunits='kcal/mol',
        color='#20a8d8'
    )

    filename = tmpdir / 'plot.html'
    fig.dump(filename)

    with open(filename, 'r') as fd:
        graph = fd.read()

    results_file = os.path.join(results_path, 'simple_plot.html')
    try:
        with open(results_file, 'r') as fd:
            correct_result = fd.read()
    except Exception:
        correct_result = ''

    if graph != correct_result:
        with open(results_file + '+', 'w') as fd:
            fd.write(graph)

    assert graph == correct_result



# -*- coding: utf-8 -*-

"""
A package for handling plotly plots in SEAMM.

A figure comprises one or more plots (or graphs) laid out in a grid.
The datasets within a plot are represetned by traces, lines, scatter
plots, pie chart elements, etc.

This package consists of a Figure class which is a container, a Plot
class that handles the actual plotting, and a Trace class for the
individual datasets and their graphical representations. The figure
class uses a grid similar to the Tk grid to control the layout of the
plots.

This package does not directly interact with plotly. Rather, it uses templates
and Jinja to create the JSON, HTML or other representation of a plotly figure.
The various containers in this package contain dictionaries of Jinja keys and
the associated values. The keys needed are defined by the template used.
"""

from .dictionary import Dictionary
import json


class Figure(Dictionary):
    """Holds one or more subplots and controls the layout.
    """

    def __init__(self, *args, jinja_env=None, template=None, **kwargs):
        """Initialize a figure.

        Keyword arguments:
        """

        super().__init__(*args, **kwargs)

        self._jinja_env = jinja_env
        self._template = template
        self._plots = Dictionary(ordered=True)

        self._grid = {'nrows': 0, 'ncolumns': 0, 'plots': {}, 'layout': {}}

    @property
    def template(self):
        """The template to use for this figure."""
        return self._template

    @template.setter
    def template(self, value):
        self._template = value

    def add_plot(self, name, **kwargs):
        """Create a new plot and return it.

        Parameters
        ----------
        name : str
            The name of the plot, which must be unique.

        Returns
        -------
        The plot object.
        """
        if name in self._plots:
            raise KeyError("Plot '{}' already exists.".format(name))

        self._plots[name] = Plot(**kwargs)

        return self._plots[name]

    def get_plot(self, name):
        """Return the named plot.

        Parameters
        ----------
        name : str
            The name of the plot, which must exist.

        Returns
        -------
        The plot object.
        """
        if name not in self._plots:
            raise KeyError("Plot '{}' does not exist.".format(name))

        return self._plots[name]

    def grid_plots(self, *args, **kwargs):
        """Define the layout of the plots.

        When there is more than one plot, they need to be laid out in a grid.
        This method is called to define the layout, and will follow closely the
        approach used in Tk grid.

        Parameters
        ----------
        *args
            Positional arguments specifying the layout of one or more rows.
            each row is given as a string composed of the following:

                <plot> :
                     is the name of a plot.

                x :
                    leaves an empty column between the plot on the
                    left and the plot on the right.

                ^ :
                    extends the row span of the plot above the ^'s in the
                    grid. The number of ^'s in a row must match the number
                    of columns spanned by the plot above it.

                - :
                    increases the column span of the plot to the
                    left. Several -'s in a row will successively increase
                    the number of columns spanned. A - may not follow a ^
                    or a x, nor may it be the first plot argument to
                    grid configure.

        **kwargs
            Keyword arguments for options for the given rows:
                padx : float, optional
                    The fraction of space to leave between plots in a row.
                    Defaults to 0.02.
                pady : float, optional
                    The fraction of space to leave between rows. Defaults to
                    0.02.
        """

        grid = self._grid
        plots = grid['plots'] = []
        # First, check that the layout is valid
        last = ''
        for row_spec in args:
            column = 0
            for key in row_spec.split():
                if column == 0:
                    if key == '-':
                        raise RuntimeError(
                            "'-' cannot be the first item in a row: '{}'"
                            .format(row_spec)
                        )
                else:
                    if key == '-' and (last == '^' or last == 'x'):
                        raise RuntimeError(
                            "'-' cannot follow 'x' or '^': '{}'"
                            .format(row_spec)
                        )
                last = key
                if key in '-^x':
                    pass
                else:
                    if key in plots:
                        raise RuntimeError(
                            r"Plot '{}' is already used: '{}'".format(
                                key, row_spec
                            )
                        )
                    plots.append(key)
                column += 1

        # Now process the entire specification
        layout = grid['layout']
        for row_spec in args:
            row = grid['nrows']
            grid['nrows'] += 1
            column = 0
            for key in row_spec.split():
                layout[(row, column)] = key
                column += 1
            if column > grid['ncolumns']:
                grid['ncolumns'] = column

        # Get row and column spans for the plots, checking that the input
        # is valid

        nrows = grid['nrows']
        ncolumns = grid['ncolumns']
        plots = grid['plots'] = []
        for row in range(nrows):
            for column in range(ncolumns):
                name = layout[(row, column)]
                if name not in '-^x':
                    plot = self.get_plot(name)
                    plot.row = row
                    plot.column = column
                    plot.column_span = 1

                    plots.append(plot)

                    # Walk to the right looking for cells with '-'
                    for tmp in range(column + 1, ncolumns):
                        if layout[(row, tmp)] == '-':
                            plot.column_span += 1
                        else:
                            break
                    plot.row_span = 1
                    for tmp in range(row + 1, nrows):
                        # Walk down looking for cell with '^'
                        if layout[(tmp, column)] == '^':
                            plot.row_span += 1

                            # If the plot spans columns and rows, all the
                            # cells in the extra rows must be '^'.
                            # Check this!
                            for col in range(
                                column + 1, column + plot.column_span
                            ):
                                if layout[(tmp, col)] != '^':
                                    val = layout[(tmp, col)]
                                    raise RuntimeError(
                                        (
                                            "The cell at {},{} should be a"
                                            " '^', not '{}'"
                                        ).format(tmp, col, val)
                                    )
                        else:
                            break

    def dump(self, filename):
        """Write the filled in template to disk as <filename>.

        Parameters
        ----------
        filename : str or filepath
            The name or path to the file to write.

        Returns
        -------
        nothing
        """

        text = self.dumps()
        with open(filename, 'w') as fd:
            fd.write(text)

    def dumps(self):
        """Return the filled in template document as a string.

        Merge the layout data and the traces into the template document and
        return a string version of the document.

        Returns
        -------
        str
            The resulting document.
        """
        # Ensure that the jsonify filter is available in Jinja

        if 'jsonify' not in self._jinja_env.filters:
            self._jinja_env.filters['jsonify'] = json.dumps

        # Get the layout (or create a default if there isn't one)
        # In the process get a list of the plots...
        plots = []
        grid = self._grid

        if len(grid['layout']) == 0:
            # Set up the default layout of a vertical column of subplots
            nrows = 0
            ncolumns = 1
            layout = {}
            for plot in self._plots.values():
                plots.append(plot)
                layout[(nrows, 0)] = plot
                plot.row = nrows
                plot.column = 0
                nrows += 1
        else:
            nrows = grid['nrows']
            ncolumns = grid['ncolumns']
            plots = grid['plots']

        # Work out where the plots (or really their axes) go
        padx = 0.1
        pady = 0.1
        width = 1.0 / ncolumns
        height = 1.0 / nrows

        for plot in plots:
            row = plot.row
            column = plot.column
            row_span = plot.row_span
            column_span = plot.column_span

            if row == 0:
                top = 1.0
            else:
                top = 1 - row * height - pady

            left = column * width
            if column + column_span == ncolumns:
                right = 1.0
            else:
                right = left + column_span * width - padx
            bottom = 1 - (row + row_span) * height

            plot.left = left
            plot.right = right
            plot.top = top
            plot.bottom = bottom

        # Number the axes sequentially and set their limits
        axis_number = 1
        for plot in plots:
            for axis in plot.axes:
                xyz = axis.direction
                axis.update(
                    {
                        'number': axis_number,
                        'name': xyz + 'axis' + str(axis_number),
                        'short_name': xyz + str(axis_number)
                    }
                )
                if xyz == 'x':
                    axis.update({'start': plot.left, 'stop': plot.right})
                elif xyz == 'y':
                    axis.update({'start': plot.bottom, 'stop': plot.top})
                axis_number += 1

        # Sort out any anchors between axes and get the data
        axes = []
        for plot in plots:
            for axis in plot.axes:
                if axis.anchor is None:
                    axis.update(anchor='free')
                else:
                    axis.update(anchor=axis.anchor['short_name'])

                axes.append(axis.to_dict())

        # And, finally, set up the traces
        traces = []
        for plot in plots:
            for trace in plot.values():
                if trace.x_axis is not None:
                    trace.update(xaxis=trace.x_axis['short_name'])
                if trace.y_axis is not None:
                    trace.update(yaxis=trace.y_axis['short_name'])
                if trace.z_axis is not None:
                    trace.update(zaxis=trace.z_axis['short_name'])

                traces.append(trace.to_dict())

        # Assemble all the data and pass to Jinja
        tmp = dict(self)
        tmp['traces'] = traces
        tmp['axes'] = axes

        template = self._jinja_env.get_template(self.template)
        result = template.render(tmp)

        return result


class Plot(Dictionary):
    """Class to represent a plotly plot.

    The Plot class contains Axes and Traces, which represent the
    datasets being plotted.  This class is a dict-like object where
    the elements of the dictionary are the traces.
    """

    def __init__(
        self,
        left=0,
        right=1,
        top=1,
        bottom=0,
        row=0,
        column=0,
        column_span=1,
        row_span=1
    ):
        """Initialize a plot, optionally with data.

        Keyword arguments:
        """
        super().__init__(ordered=True)

        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom
        self.row = row
        self.column = column
        self.column_span = column_span
        self.row_span = row_span

        self._axes = []

    @property
    def axes(self):
        """The list of axes for this plot."""
        return self._axes

    def add_axis(self, direction, anchor=None, **kwargs):
        """Create a new axis and return it.

        Parameters
        ----------
        direction : str (one of 'x', 'y', or 'z')
            The direction of this axis.

        Returns
        -------
        The axis object.
        """
        axis = Axis(direction, anchor=anchor, **kwargs)
        self._axes.append(axis)
        return axis

    def add_trace(self, name, x_axis=None, y_axis=None, z_axis=None, **kwargs):
        """Create a new trace and return it.

        Parameters
        ----------
        name : str
            The name of the trace, which must be unique.

        Returns
        -------
        The trace object.
        """
        if name in self:
            raise KeyError("Trace '{}' already exists.".format(name))

        self[name] = Trace(
            x_axis=x_axis, y_axis=y_axis, z_axis=z_axis, name=name, **kwargs
        )

        return self[name]


class Axis(Dictionary):
    """Class to represent an axis in a plotly plot.

    The Axis class defines an axis and the  associated graphical
    control for how to display them. The class is a dict-like object
    where the elements of the dictionary are the control parameters
    for the axis.
    """

    def __init__(self, direction, *args, anchor=None, **kwargs):
        """Initialize an axis, optionally with data.

        Parameters
        ----------
        direction : str (one of 'x', 'y', or 'z')
            The direction for this axis

        args : dict
            0+ dictionaries to update from, in order

        kwargs : keyword - value pairs
            0+ named arguments to update from
        """
        super().__init__(ordered=True)

        self.anchor = anchor
        self.direction = direction
        self['label'] = ''
        self['number'] = None
        self['start'] = 0.0
        self['stop'] = 1.0

        self.update(*args, **kwargs)

    def to_dict(self):
        """Return a dictionary representing the axis.

        Returns
        -------
        dict
            A dictionary of the axis data, suitable for Jinja.
        """
        return dict(self.data)


class Trace(Dictionary):
    """Class to represent a trace in a plotly plot.

    The Trace class contains the datasets and associated graphical
    control for how to display them. The class is a dict-like object
    where the elements of the dictionary are the traces.
    """

    def __init__(self, x_axis=None, y_axis=None, z_axis=None, **kwargs):
        """Initialize a trace, optionally with data.

        Keyword arguments:
        args : dict
            0+ dictionaries to update from, in order
        kwargs : keyword - value pairs
            0+ named arguments to update from
        """
        super().__init__(**kwargs)

        self.x_axis = x_axis
        self.y_axis = y_axis
        self.z_axis = z_axis

    def to_dict(self):
        """Return a dictionary representing the trace.

        Returns
        -------
        dict
            A dictionary of the trace data, suitable for Jinja.
        """
        return dict(self.data)



# -*- coding: utf-8 -*-

"""Utility routines to help with variable names for scripts"""

from keyword import iskeyword
import re


def is_valid(name):
    """Check if a variable name is valid and is not a keyword"""
    return name.isidentifier() and not iskeyword(name)


def clean(raw_name):
    """Fix up an input string so that it is a valid variable name"""

    # Remove invalid characters
    name = re.sub('[^0-9a-zA-Z_]', '_', raw_name)

    # Check that it is a valid variable name. If not
    # put an underscore on the front

    if not is_valid(name):
        name = '_' + name

    # should be valid, but check once more
    if not is_valid(name):
        raise RuntimeError(
            'Variable name {} (from {}) is not valid!'.format(name, raw_name)
        )

    return name



# -*- coding: utf-8 -*-

"""A class for wrapping output from the MolSSI Framework and plugins.
Initially it is a thin wrapper around a dictionary to store the output,
using JSON to serialize and deserialize the data.

It presents the dict api with the addition of methods to serialize and
deserialize the contents.
"""

import bz2
import gzip
import logging
import os.path

logger = logging.getLogger(__name__)


class File():

    def __init__(
        self,
        filename,
        mode,
        compression=None,
        organization='MolSSI',
        filetype=None,
        version=None
    ):
        self.filename = filename
        self.mode = mode

        # Depending on the mode, may require filetype and version
        if self.mode == 'w':
            if filetype is None:
                raise ValueError("Filetype must be given for writable files!")
            if version is None:
                raise ValueError("Version must be given for writable files!")
        elif self.mode == 'a':
            if os.path.isfile(self.filename):
                # Get the header and decipher
                fd = self.__enter__()
                try:
                    line = next(fd)
                except:  # noqa: E722
                    raise RuntimeError("problem reading header line!")

                fd.close()

                # if line[0] == '!' and len(line.split()) == 3:
                #     organization, filetype, version = line[1:].split()
            else:
                logger.warning(
                    "reading '{}', expected a header line but got\n\t'{}'"
                    .format(self.filename, line)
                )

                if filetype is None:
                    raise ValueError(
                        "Filetype must be given to create a file!"
                    )
                if version is None:
                    raise ValueError("Version must be given to create file!")

        # Determine the type of the file
        if compression is not None:
            if compression not in ('text', 'bz2', 'gzip'):
                raise ValueError(
                    "Invalid compression: '{}'".format(compression)
                )
        else:
            extension = os.path.splitext(self.filename)[1].strip().lower()
            if extension == '.bz2':
                compression = 'bzip2'
            elif extension == '.gz':
                compression = 'gzip'
            else:
                compression = 'text'
        self.compression = compression

    def __enter__(self):
        if self.compression == 'bzip2':
            self.file_descriptor = bz2.open(self.filename, self.mode + 't')
        elif self.compression == 'gzip':
            self.file_descriptor = gzip.open(self.filename, self.mode + 't')
        else:
            self.file_descriptor = open(self.filename, self.mode)

        return self.file_descriptor

    def __exit__(self, *args):
        self.file_descriptor.close()

    def read_header(self):
        if self.compression == 'bzip2':
            with bz2.open(self.filename, 'rt') as fd:
                line = next(fd)
        elif self.compression == 'gzip':
            with gzip.open(self.filename, 'rt') as fd:
                line = next(fd)
        else:
            with open(self.filename, 'r'):
                line = next(fd)

        if line[0] == '!' and len(line.split()) == 3:
            organization, filetype, version = line[1:].split()
            logger.info(
                "reading '{}', a {} file from {}, version {}".format(
                    self.filename, filetype, organization, version
                )
            )
        else:
            raise RuntimeError(
                "reading '{}', expected a header line but got\n\t'{}'".format(
                    self.filename, line
                )
            )



"""Tabulated data about the elements."""

# From https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020
# With missing masses from WebElements
element_data = {
    "H": {
        "atomic number": 1,
        "atomic symbol": "H",
        "element": "hydrogen",
        "atomic weight": 1.008,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "He": {
        "atomic number": 2,
        "atomic symbol": "He",
        "element": "helium",
        "atomic weight": 4.0026,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Li": {
        "atomic number": 3,
        "atomic symbol": "Li",
        "element": "lithium",
        "atomic weight": 6.967499999999999,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Be": {
        "atomic number": 4,
        "atomic symbol": "Be",
        "element": "beryllium",
        "atomic weight": 9.0122,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "B": {
        "atomic number": 5,
        "atomic symbol": "B",
        "element": "boron",
        "atomic weight": 10.8135,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "C": {
        "atomic number": 6,
        "atomic symbol": "C",
        "element": "carbon",
        "atomic weight": 12.0105,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "N": {
        "atomic number": 7,
        "atomic symbol": "N",
        "element": "nitrogen",
        "atomic weight": 14.007,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "O": {
        "atomic number": 8,
        "atomic symbol": "O",
        "element": "oxygen",
        "atomic weight": 15.999500000000001,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "F": {
        "atomic number": 9,
        "atomic symbol": "F",
        "element": "fluorine",
        "atomic weight": 18.998,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ne": {
        "atomic number": 10,
        "atomic symbol": "Ne",
        "element": "neon",
        "atomic weight": 20.18,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Na": {
        "atomic number": 11,
        "atomic symbol": "Na",
        "element": "sodium",
        "atomic weight": 22.99,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Mg": {
        "atomic number": 12,
        "atomic symbol": "Mg",
        "element": "magnesium",
        "atomic weight": 24.3055,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Al": {
        "atomic number": 13,
        "atomic symbol": "Al",
        "element": "aluminium",
        "atomic weight": 26.982,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Si": {
        "atomic number": 14,
        "atomic symbol": "Si",
        "element": "silicon",
        "atomic weight": 28.085,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "P": {
        "atomic number": 15,
        "atomic symbol": "P",
        "element": "phosphorus",
        "atomic weight": 30.974,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "S": {
        "atomic number": 16,
        "atomic symbol": "S",
        "element": "sulfur",
        "atomic weight": 32.067499999999995,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Cl": {
        "atomic number": 17,
        "atomic symbol": "Cl",
        "element": "chlorine",
        "atomic weight": 35.451499999999996,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ar": {
        "atomic number": 18,
        "atomic symbol": "Ar",
        "element": "argon",
        "atomic weight": 39.8775,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "K": {
        "atomic number": 19,
        "atomic symbol": "K",
        "element": "potassium",
        "atomic weight": 39.098,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ca": {
        "atomic number": 20,
        "atomic symbol": "Ca",
        "element": "calcium",
        "atomic weight": 40.078,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Sc": {
        "atomic number": 21,
        "atomic symbol": "Sc",
        "element": "scandium",
        "atomic weight": 44.956,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ti": {
        "atomic number": 22,
        "atomic symbol": "Ti",
        "element": "titanium",
        "atomic weight": 47.867,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "V": {
        "atomic number": 23,
        "atomic symbol": "V",
        "element": "vanadium",
        "atomic weight": 50.942,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Cr": {
        "atomic number": 24,
        "atomic symbol": "Cr",
        "element": "chromium",
        "atomic weight": 51.996,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Mn": {
        "atomic number": 25,
        "atomic symbol": "Mn",
        "element": "manganese",
        "atomic weight": 54.938,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Fe": {
        "atomic number": 26,
        "atomic symbol": "Fe",
        "element": "iron",
        "atomic weight": 55.845,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Co": {
        "atomic number": 27,
        "atomic symbol": "Co",
        "element": "cobalt",
        "atomic weight": 58.933,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ni": {
        "atomic number": 28,
        "atomic symbol": "Ni",
        "element": "nickel",
        "atomic weight": 58.693,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Cu": {
        "atomic number": 29,
        "atomic symbol": "Cu",
        "element": "copper",
        "atomic weight": 63.546,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Zn": {
        "atomic number": 30,
        "atomic symbol": "Zn",
        "element": "zinc",
        "atomic weight": 65.38,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ga": {
        "atomic number": 31,
        "atomic symbol": "Ga",
        "element": "gallium",
        "atomic weight": 69.723,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ge": {
        "atomic number": 32,
        "atomic symbol": "Ge",
        "element": "germanium",
        "atomic weight": 72.63,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "As": {
        "atomic number": 33,
        "atomic symbol": "As",
        "element": "arsenic",
        "atomic weight": 74.922,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Se": {
        "atomic number": 34,
        "atomic symbol": "Se",
        "element": "selenium",
        "atomic weight": 78.971,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Br": {
        "atomic number": 35,
        "atomic symbol": "Br",
        "element": "bromine",
        "atomic weight": 79.904,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Kr": {
        "atomic number": 36,
        "atomic symbol": "Kr",
        "element": "krypton",
        "atomic weight": 83.798,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Rb": {
        "atomic number": 37,
        "atomic symbol": "Rb",
        "element": "rubidium",
        "atomic weight": 85.468,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Sr": {
        "atomic number": 38,
        "atomic symbol": "Sr",
        "element": "strontium",
        "atomic weight": 87.62,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Y": {
        "atomic number": 39,
        "atomic symbol": "Y",
        "element": "yttrium",
        "atomic weight": 88.906,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Zr": {
        "atomic number": 40,
        "atomic symbol": "Zr",
        "element": "zirconium",
        "atomic weight": 91.224,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Nb": {
        "atomic number": 41,
        "atomic symbol": "Nb",
        "element": "niobium",
        "atomic weight": 92.906,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Mo": {
        "atomic number": 42,
        "atomic symbol": "Mo",
        "element": "molybdenum",
        "atomic weight": 95.95,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Tc": {
        "atomic number": 43,
        "atomic symbol": "Tc",
        "element": "technetium",
        "atomic weight": 96.906,
        "source": "webelements  30 July 2020"
    },
    "Ru": {
        "atomic number": 44,
        "atomic symbol": "Ru",
        "element": "ruthenium",
        "atomic weight": 101.07,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Rh": {
        "atomic number": 45,
        "atomic symbol": "Rh",
        "element": "rhodium",
        "atomic weight": 102.91,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Pd": {
        "atomic number": 46,
        "atomic symbol": "Pd",
        "element": "palladium",
        "atomic weight": 106.42,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ag": {
        "atomic number": 47,
        "atomic symbol": "Ag",
        "element": "silver",
        "atomic weight": 107.87,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Cd": {
        "atomic number": 48,
        "atomic symbol": "Cd",
        "element": "cadmium",
        "atomic weight": 112.41,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "In": {
        "atomic number": 49,
        "atomic symbol": "In",
        "element": "indium",
        "atomic weight": 114.82,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Sn": {
        "atomic number": 50,
        "atomic symbol": "Sn",
        "element": "tin",
        "atomic weight": 118.71,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Sb": {
        "atomic number": 51,
        "atomic symbol": "Sb",
        "element": "antimony",
        "atomic weight": 121.76,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Te": {
        "atomic number": 52,
        "atomic symbol": "Te",
        "element": "tellurium",
        "atomic weight": 127.6,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "I": {
        "atomic number": 53,
        "atomic symbol": "I",
        "element": "iodine",
        "atomic weight": 126.9,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Xe": {
        "atomic number": 54,
        "atomic symbol": "Xe",
        "element": "xenon",
        "atomic weight": 131.29,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Cs": {
        "atomic number": 55,
        "atomic symbol": "Cs",
        "element": "caesium",
        "atomic weight": 132.91,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ba": {
        "atomic number": 56,
        "atomic symbol": "Ba",
        "element": "barium",
        "atomic weight": 137.33,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "La": {
        "atomic number": 57,
        "atomic symbol": "La",
        "element": "lanthanum",
        "atomic weight": 138.91,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ce": {
        "atomic number": 58,
        "atomic symbol": "Ce",
        "element": "cerium",
        "atomic weight": 140.12,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Pr": {
        "atomic number": 59,
        "atomic symbol": "Pr",
        "element": "praseodymium",
        "atomic weight": 140.91,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Nd": {
        "atomic number": 60,
        "atomic symbol": "Nd",
        "element": "neodymium",
        "atomic weight": 144.24,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Pm": {
        "atomic number": 61,
        "atomic symbol": "Pm",
        "element": "promethium",
        "atomic weight": 144.91,
        "source": "webelements  30 July 2020"
    },
    "Sm": {
        "atomic number": 62,
        "atomic symbol": "Sm",
        "element": "samarium",
        "atomic weight": 150.36,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Eu": {
        "atomic number": 63,
        "atomic symbol": "Eu",
        "element": "europium",
        "atomic weight": 151.96,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Gd": {
        "atomic number": 64,
        "atomic symbol": "Gd",
        "element": "gadolinium",
        "atomic weight": 157.25,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Tb": {
        "atomic number": 65,
        "atomic symbol": "Tb",
        "element": "terbium",
        "atomic weight": 158.93,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Dy": {
        "atomic number": 66,
        "atomic symbol": "Dy",
        "element": "dysprosium",
        "atomic weight": 162.5,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ho": {
        "atomic number": 67,
        "atomic symbol": "Ho",
        "element": "holmium",
        "atomic weight": 164.93,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Er": {
        "atomic number": 68,
        "atomic symbol": "Er",
        "element": "erbium",
        "atomic weight": 167.26,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Tm": {
        "atomic number": 69,
        "atomic symbol": "Tm",
        "element": "thulium",
        "atomic weight": 168.93,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Yb": {
        "atomic number": 70,
        "atomic symbol": "Yb",
        "element": "ytterbium",
        "atomic weight": 173.05,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Lu": {
        "atomic number": 71,
        "atomic symbol": "Lu",
        "element": "lutetium",
        "atomic weight": 174.97,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Hf": {
        "atomic number": 72,
        "atomic symbol": "Hf",
        "element": "hafnium",
        "atomic weight": 178.49,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ta": {
        "atomic number": 73,
        "atomic symbol": "Ta",
        "element": "tantalum",
        "atomic weight": 180.95,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "W": {
        "atomic number": 74,
        "atomic symbol": "W",
        "element": "tungsten",
        "atomic weight": 183.84,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Re": {
        "atomic number": 75,
        "atomic symbol": "Re",
        "element": "rhenium",
        "atomic weight": 186.21,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Os": {
        "atomic number": 76,
        "atomic symbol": "Os",
        "element": "osmium",
        "atomic weight": 190.23,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Ir": {
        "atomic number": 77,
        "atomic symbol": "Ir",
        "element": "iridium",
        "atomic weight": 192.22,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Pt": {
        "atomic number": 78,
        "atomic symbol": "Pt",
        "element": "platinum",
        "atomic weight": 195.08,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Au": {
        "atomic number": 79,
        "atomic symbol": "Au",
        "element": "gold",
        "atomic weight": 196.97,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Hg": {
        "atomic number": 80,
        "atomic symbol": "Hg",
        "element": "mercury",
        "atomic weight": 200.59,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Tl": {
        "atomic number": 81,
        "atomic symbol": "Tl",
        "element": "thallium",
        "atomic weight": 204.385,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Pb": {
        "atomic number": 82,
        "atomic symbol": "Pb",
        "element": "lead",
        "atomic weight": 207.2,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Bi": {
        "atomic number": 83,
        "atomic symbol": "Bi",
        "element": "bismuth",
        "atomic weight": 208.98,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Po": {
        "atomic number": 84,
        "atomic symbol": "Po",
        "element": "polonium",
        "atomic weight": 208.98,
        "source": "webelements  30 July 2020"
    },
    "At": {
        "atomic number": 85,
        "atomic symbol": "At",
        "element": "astatine",
        "atomic weight": 209.99,
        "source": "webelements  30 July 2020"
    },
    "Rn": {
        "atomic number": 86,
        "atomic symbol": "Rn",
        "element": "radon",
        "atomic weight": 222.02,
        "source": "webelements  30 July 2020"
    },
    "Fr": {
        "atomic number": 87,
        "atomic symbol": "Fr",
        "element": "francium",
        "atomic weight": 223.02,
        "source": "webelements  30 July 2020"
    },
    "Ra": {
        "atomic number": 88,
        "atomic symbol": "Ra",
        "element": "radium",
        "atomic weight": 226.03,
        "source": "webelements  30 July 2020"
    },
    "Ac": {
        "atomic number": 89,
        "atomic symbol": "Ac",
        "element": "actinium",
        "atomic weight": 227.03,
        "source": "webelements  30 July 2020"
    },
    "Th": {
        "atomic number": 90,
        "atomic symbol": "Th",
        "element": "thorium",
        "atomic weight": 232.04,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Pa": {
        "atomic number": 91,
        "atomic symbol": "Pa",
        "element": "protactinium",
        "atomic weight": 231.04,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "U": {
        "atomic number": 92,
        "atomic symbol": "U",
        "element": "uranium",
        "atomic weight": 238.03,
        "source": "https://ciaaw.org/abridged-atomic-weights.htm  30 July 2020"
    },
    "Np": {
        "atomic number": 93,
        "atomic symbol": "Np",
        "element": "neptunium",
        "atomic weight": 237.05,
        "source": "webelements  30 July 2020"
    },
    "Pu": {
        "atomic number": 94,
        "atomic symbol": "Pu",
        "element": "plutonium",
        "atomic weight": 244.06,
        "source": "webelements  30 July 2020"
    },
    "Am": {
        "atomic number": 95,
        "atomic symbol": "Am",
        "element": "americium",
        "atomic weight": 243.06,
        "source": "webelements  30 July 2020"
    },
    "Cm": {
        "atomic number": 96,
        "atomic symbol": "Cm",
        "element": "curium",
        "atomic weight": 247.07,
        "source": "webelements  30 July 2020"
    },
    "Bk": {
        "atomic number": 97,
        "atomic symbol": "Bk",
        "element": "berkelium",
        "atomic weight": 247.07,
        "source": "webelements  30 July 2020"
    },
    "Cf": {
        "atomic number": 98,
        "atomic symbol": "Cf",
        "element": "californium",
        "atomic weight": 251.08,
        "source": "webelements  30 July 2020"
    },
    "Es": {
        "atomic number": 99,
        "atomic symbol": "Es",
        "element": "einsteinium",
        "atomic weight": 252.08,
        "source": "webelements  30 July 2020"
    },
    "Fm": {
        "atomic number": 100,
        "atomic symbol": "Fm",
        "element": "fermium",
        "atomic weight": 257.10,
        "source": "webelements  30 July 2020"
    },
    "Md": {
        "atomic number": 101,
        "atomic symbol": "Md",
        "element": "mendelevium",
        "atomic weight": 258.10,
        "source": "webelements  30 July 2020"
    },
    "No": {
        "atomic number": 102,
        "atomic symbol": "No",
        "element": "nobelium",
        "atomic weight": 259.10,
        "source": "webelements  30 July 2020"
    },
    "Lr": {
        "atomic number": 103,
        "atomic symbol": "Lr",
        "element": "lawrencium",
        "atomic weight": 262.11,
        "source": "webelements  30 July 2020"
    },
    "Rf": {
        "atomic number": 104,
        "atomic symbol": "Rf",
        "element": "rutherfordium",
        "atomic weight": 267.12,
        "source": "webelements  30 July 2020"
    },
    "Db": {
        "atomic number": 105,
        "atomic symbol": "Db",
        "element": "dubnium",
        "atomic weight": 270.13,
        "source": "webelements  30 July 2020"
    },
    "Sg": {
        "atomic number": 106,
        "atomic symbol": "Sg",
        "element": "seaborgium",
        "atomic weight": 269.13,
        "source": "webelements  30 July 2020"
    },
    "Bh": {
        "atomic number": 107,
        "atomic symbol": "Bh",
        "element": "bohrium",
        "atomic weight": 270.13,
        "source": "webelements  30 July 2020"
    },
    "Hs": {
        "atomic number": 108,
        "atomic symbol": "Hs",
        "element": "hassium",
        "atomic weight": 269.13,
        "source": "webelements  30 July 2020"
    },
    "Mt": {
        "atomic number": 109,
        "atomic symbol": "Mt",
        "element": "meitnerium",
        "atomic weight": 278.16,
        "source": "webelements  30 July 2020"
    },
    "Ds": {
        "atomic number": 110,
        "atomic symbol": "Ds",
        "element": "darmstadtium",
        "atomic weight": 281.17,
        "source": "webelements  30 July 2020"
    },
    "Rg": {
        "atomic number": 111,
        "atomic symbol": "Rg",
        "element": "roentgenium",
        "atomic weight": 281.17,
        "source": "webelements  30 July 2020"
    },
    "Cn": {
        "atomic number": 112,
        "atomic symbol": "Cn",
        "element": "copernicium",
        "atomic weight": 285.18,
        "source": "webelements  30 July 2020"
    },
    "Nh": {
        "atomic number": 113,
        "atomic symbol": "Nh",
        "element": "nihonium",
        "atomic weight": 286.18,
        "source": "webelements  30 July 2020"
    },
    "Fl": {
        "atomic number": 114,
        "atomic symbol": "Fl",
        "element": "flerovium",
        "atomic weight": 289.19,
        "source": "webelements  30 July 2020"
    },
    "Mc": {
        "atomic number": 115,
        "atomic symbol": "Mc",
        "element": "moscovium",
        "atomic weight": 289.20,
        "source": "webelements  30 July 2020"
    },
    "Lv": {
        "atomic number": 116,
        "atomic symbol": "Lv",
        "element": "livermorium",
        "atomic weight": 293.20,
        "source": "webelements  30 July 2020"
    },
    "Ts": {
        "atomic number": 117,
        "atomic symbol": "Ts",
        "element": "tennessine",
        "atomic weight": 293.21,
        "source": "webelements  30 July 2020"
    },
    "Og": {
        "atomic number": 118,
        "atomic symbol": "Og",
        "element": "oganesson",
        "atomic weight": 294.21,
        "source": "webelements  30 July 2020"
    }
}



# -*- coding: utf-8 -*-

"""
seamm_util
Utility functions for the SEAMM environment.
"""

# Bring up the classes so that they appear to be directly in
# the packmol_step package.

from seamm_util.elemental_data import element_data  # noqa: F401
from seamm_util.check_executable import check_executable  # noqa: F401
from seamm_util.dictionary import Dictionary  # noqa: F401
from seamm_util.units import ureg, Q_, units_class  # noqa: F401
from seamm_util.include_open import Open  # noqa: F401
from seamm_util.include_open import splitext  # noqa: F401
from seamm_util.seamm_json import JSONDecoder  # noqa: F401
from seamm_util.seamm_json import JSONEncoder  # noqa: F401
from seamm_util.plotting import Figure  # noqa: F401
import seamm_util.printing  # noqa: F401
import seamm_util.variable_names  # noqa: F401
import seamm_util.water_models  # noqa: F401

# Handle versioneer
from ._version import get_versions
__author__ = """Paul Saxe"""
__email__ = 'psaxe@molssi.org'
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



"""
The Printer class provides control over printing output, based
on the Python logging module. The levels a translated to more reasonable
names for output, rather than debugging:

        Level        Logger Level    Numeric value
        --------     ------------    -------------
        JOB          CRITICAL	           50
        IMPORTANT    ERROR	           40
        TERSE        WARNING	           30
        NORMAL       INFO                  20
        VERBOSE      DEBUG                 10
        ALL          ALL                    0

By default, JOB level output is mirrored in the main log for the job
as well as the local log for the stage. All output from the cutoff level
up, which is NORMAL by default, is placed in the local stage log. Optionally,
output from a cutoff level and up is also written to the standard output.
"""

import inspect
import logging
import os
import sys
import textwrap
try:
    import thread
    import threading
except ImportError:
    thread = None

# ---------------------------------------------------------------------------
#   Level related stuff
# ---------------------------------------------------------------------------
#
# Default levels and level names, these can be replaced with any positive set
# of values having corresponding names. There is a pseudo-level, ALL, which
# is only really there as a lower limit for user-defined levels. Handlers and
# loggers are initialized with ALL so that they will log all messages, even
# at user-defined levels.
#

JOB = 50
IMPORTANT = 40
TERSE = 30
NORMAL = 20
VERBOSE = 10
ALL = 0

_levelNames = {
    JOB: 'JOB',
    IMPORTANT: 'IMPORTANT',
    TERSE: 'TERSE',
    NORMAL: 'NORMAL',
    VERBOSE: 'VERBOSE',
    ALL: 'ALL',
    'JOB': JOB,
    'IMPORTANT': IMPORTANT,
    'TERSE': TERSE,
    'NORMAL': NORMAL,
    'VERBOSE': VERBOSE,
    'ALL': ALL,
}


def getLevelName(lvl):
    """
    Return the textual representation of printing level 'lvl'. If the level is
    one of the predefined levels (JOB, IMPORTANT, TERSE, NORMAL, VERBOSE) then
    you get the corresponding string. If you have associated levels with names
    using addLevelName then the name you have associated with 'lvl' is
    returned. Otherwise, the string "Level %s" % lvl is returned.
    """
    return _levelNames.get(lvl, "Level {}".format(lvl))


def addLevelName(lvl, levelName):
    """
    Associate 'levelName' with 'lvl'. This is used when converting levels
    to text during message formatting.
    """
    _acquireLock()
    try:  # unlikely to cause an exception, but you never know...
        _levelNames[lvl] = levelName
        _levelNames[levelName] = lvl
    finally:
        _releaseLock()


# ---------------------------------------------------------------------------
#   Thread-related stuff
# ---------------------------------------------------------------------------

#
# _lock is used to serialize access to shared data structures in this module.
# This needs to be an RLock because fileConfig() creates Handlers and so
# might arbitrary user threads. Since Handler.__init__() updates the shared
# dictionary _handlers, it needs to acquire the lock. But if configuring,
# the lock would already have been acquired - so we need an RLock.
# The same argument applies to Loggers and Manager.loggerDict.
#

_lock = None


def _acquireLock():
    """
    Acquire the module-level lock for serializing access to shared data.
    This should be released with _releaseLock().
    """
    global _lock
    if (not _lock) and thread:
        _lock = threading.RLock()
    if _lock:
        _lock.acquire()


def _releaseLock():
    """
    Release the module-level lock acquired by calling _acquireLock().
    """
    if _lock:
        _lock.release()


# ---------------------------------------------------------------------------
#   Manager classes and functions
# ---------------------------------------------------------------------------


class PlaceHolder(object):
    """
    PlaceHolder instances are used in the Manager printer hierarchy to take
    the place of nodes for which no printers have been defined [FIXME add
    example].
    """

    def __init__(self, aprinter):
        """
        Initialize with the specified printer being a child of this
        placeholder.
        """
        self.printers = [aprinter]

    def append(self, aprinter):
        """
        Add the specified printer as a child of this placeholder.
        """
        if aprinter not in self.printers:
            self.printers.append(aprinter)


#
#   Determine which class to use when instantiating printers.
#
_printerClass = None


def setPrinterClass(klass):
    """
    Set the class to be used when instantiating a printer. The class should
    define __init__() such that only a name argument is required, and the
    __init__() should call Printer.__init__()
    """
    if klass != Printer:
        if not isinstance(klass, object):
            raise TypeError("setPrinterClass is expecting a class")
        if not issubclass(klass, Printer):
            raise TypeError(
                "printer not derived from printer.Printer: " + klass.__name__
            )
    global _printerClass
    _printerClass = klass


class Manager(object):
    """
    There is [under normal circumstances] just one Manager instance, which
    holds the hierarchy of printers.
    """

    def __init__(self, print_root):
        """
        Initialize the manager with the root node of the printer hierarchy.
        """
        self.print_root = print_root
        self.disable = 0
        self.emittedNoHandlerWarning = 0
        self.printerDict = {}

    def getPrinter(self, name):
        """
        Get a printer with the specified name (channel name), creating it
        if it doesn't yet exist. If a PlaceHolder existed for the specified
        name [i.e. the printer didn't exist but a child of it did], replace
        it with the created printer and fix up the parent/child references
        which pointed to the placeholder to now point to the printer.
        """
        rv = None
        _acquireLock()
        try:
            if name in self.printerDict:
                rv = self.printerDict[name]
                if isinstance(rv, PlaceHolder):
                    ph = rv
                    rv = _printerClass(name)
                    rv.manager = self
                    self.printerDict[name] = rv
                    self._fixupChildren(ph, rv)
                    self._fixupParents(rv)
            else:
                rv = _printerClass(name)
                rv.manager = self
                self.printerDict[name] = rv
                self._fixupParents(rv)
        finally:
            _releaseLock()
        return rv

    def _fixupParents(self, aprinter):
        """
        Ensure that there are either printers or placeholders all the way
        from the specified printer to the root of the printer hierarchy.
        """
        name = aprinter.name
        i = name.rfind(".")
        rv = None
        while (i > 0) and not rv:
            substr = name[:i]
            if substr not in self.printerDict:
                self.printerDict[substr] = PlaceHolder(aprinter)
            else:
                obj = self.printerDict[substr]
                if isinstance(obj, Printer):
                    rv = obj
                else:
                    assert isinstance(obj, PlaceHolder)
                    obj.append(aprinter)
            i = name.rfind(".", 0, i - 1)
        if not rv:
            rv = self.print_root
        aprinter.parent = rv

    def _fixupChildren(self, ph, aprinter):
        """
        Ensure that children of the placeholder ph are connected to the
        specified printer.
        """
        for c in ph.printers:
            if aprinter.name in c.parent.name:
                aprinter.parent = c.parent
                c.parent = aprinter


class Printer(logging.Filterer):
    """Provide controlled printing for plugins for the MolSSI
    Framework.
    """

    def __init__(self, name, level=ALL):
        """Create the Printer object
        """

        super().__init__()

        if name == 'print_root':
            self.logger = logging.getLogger(name)
        else:
            self.logger = logging.getLogger('print_root.' + name)
        self.logger.setLevel(level)

    @property
    def name(self):
        return self.logger.name

    @name.setter
    def name(self, value):
        self.logger.name = value

    @property
    def level(self):
        return self.logger.level

    @property
    def parent(self):
        return self.logger.parent

    @parent.setter
    def parent(self, value):
        self.logger.parent = value

    @property
    def propagate(self):
        return self.logger.propagate

    @propagate.setter
    def propagate(self, value):
        self.logger.propagate = value

    @property
    def handlers(self):
        return self.logger.handlers

    @property
    def disabled(self):
        return self.logger.disabled

    @disabled.setter
    def disabled(self, value):
        self.logger.disabled = value

    def setLevel(self, lvl):
        """
        Set the printing level of this printer.
        """
        self.logger.level = lvl

    def verbose(self, msg, *args, **kwargs):
        """Logs a message with level VERBOSE on this printer.
        The msg is the message format string, and the args are the arguments
        which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)
        """
        self.print(10, msg, *args, **kwargs)

    def normal(self, msg, *args, **kwargs):
        """Logs a message with level NORMAL on this printer.
        The msg is the message format string, and the args are the arguments
        which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)
        """
        self.print(20, msg, *args, **kwargs)

    def terse(self, msg, *args, **kwargs):
        """Logs a message with level TERSE on this printer.
        The msg is the message format string, and the args are the arguments
        which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)
        """
        self.print(30, msg, *args, **kwargs)

    def important(self, msg, *args, **kwargs):
        """Logs a message with level IMPORTANT on this printer.
        The msg is the message format string, and the args are the arguments
        which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)
        """
        self.print(40, msg, *args, **kwargs)

    def job(self, msg, *args, **kwargs):
        """Logs a message with level JOB on this printer.
        The msg is the message format string, and the args are the arguments
        which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)
        """
        self.print(50, msg, *args, **kwargs)

    def print(self, lvl, msg, *args, **kwargs):
        """Prints a message with integer level lvl on this logger.
        The other arguments are interpreted as for verbose().
        """
        # print('print: lvl={}, msg={}, args={}, kwargs={}'.format(
        #     lvl, msg, args, kwargs
        # ))
        self.logger.log(lvl, msg, *args, **kwargs)

    def findCaller(self):
        """
        Find the stack frame of the caller so that we can note the source
        file name and line number.
        """
        rv = (None, None)
        frame = inspect.currentframe().f_back
        while frame:
            sfn = inspect.getsourcefile(frame)
            if sfn:
                sfn = os.path.normcase(sfn)
            if sfn != _srcfile:  # noqa: F821
                # print frame.f_code.co_code
                lineno = inspect.getlineno(frame)
                rv = (sfn, lineno)
                break
            frame = frame.f_back
        return rv

    def makeRecord(self, name, lvl, fn, lno, msg, args, exc_info):
        """
        A factory method which can be overridden in subclasses to create
        specialized LogRecords.
        """
        return self.logger.makeRecord(name, lvl, fn, lno, msg, args, exc_info)

    def _log(self, lvl, msg, args, exc_info=None):
        """
        Low-level logging routine which creates a LogRecord and then calls
        all the handlers of this logger to handle the record.
        """
        self.logger._log(self, lvl, msg, args, exc_info)

    def handle(self, record):
        """
        Call the handlers for the specified record. This method is used for
        unpickled records received from a socket, as well as those created
        locally. Logger-level filtering is applied.
        """
        self.logger.handle(record)

    def addHandler(self, hdlr):
        """
        Add the specified handler to this printer.
        """
        self.logger.addHandler(hdlr)

    def removeHandler(self, hdlr):
        """
        Remove the specified handler from this printer.
        """
        self.logger.removeHandler(hdlr)

    def callHandlers(self, record):
        """
        Loop through all handlers for this printer and its parents in the
        printer hierarchy. If no handler was found, output a one-off error
        message to sys.stderr. Stop searching up the hierarchy whenever a
        printer with the "propagate" attribute set to zero is found - that
        will be the last printer whose handlers are called.
        """
        self.logger.callHandlers(record)

    def getEffectiveLevel(self):
        """
        Loop through this printer and its parents in the printer hierarchy,
        looking for a non-zero logging level. Return the first one found.
        """
        return self.logger.getEffectiveLevel()

    def isEnabledFor(self, lvl):
        """
        Is this printer enabled for level lvl?
        """
        return self.logger.isEnabledFor(lvl)


class RootPrinter(Printer):
    """
    A root printer is not that different to any other printer, except that
    it must have a printing level and there is only one instance of it in
    the hierarchy.
    """

    def __init__(self, lvl):
        """
        Initialize the printer with the name "print_root".
        """
        super().__init__("print_root", lvl)

        # Turn of propagation and set up handler
        # so that we are separate from the loggers
        self.logger.propagate = False


_printerClass = Printer

print_root = RootPrinter(NORMAL)
Printer.print_root = print_root
Printer.manager = Manager(Printer.print_root)

# ---------------------------------------------------------------------------
# Configuration classes and functions
# ---------------------------------------------------------------------------

BASIC_FORMAT = "{message:s}"


def basicConfig():
    """
    Do basic configuration for the printing system by creating a
    StreamHandler with a default Formatter and adding it to the
    root logger.
    """
    if len(print_root.handlers) == 0:
        hdlr = logging.StreamHandler()
        fmt = logging.Formatter(BASIC_FORMAT, style='{')
        hdlr.setFormatter(fmt)
        print_root.addHandler(hdlr)


_handlers = {}  # repository of handlers (for flushing when shutdown called)


def fileConfig(fname):
    """
    Read the printing configuration from a ConfigParser-format file. This can
    be called several times from an application, allowing an end user the
    ability to select from various pre-canned configurations (if the
    developer provides a mechanism to present the choices and load the chosen
    configuration).
    In versions of ConfigParser which have the readfp method [typically
    shipped in 2.x versions of Python], you can pass in a file-like object
    rather than a filename, in which case the file-like object will be read
    using readfp.
    """
    import ConfigParser

    cp = ConfigParser.ConfigParser()
    if hasattr(cp, 'readfp') and hasattr(fname, 'read'):
        cp.readfp(fname)
    else:
        cp.read(fname)
    # first, do the formatters...
    flist = cp.get("formatters", "keys")
    if len(flist):
        flist = flist.split(",")
        formatters = {}
        for form in flist:
            sectname = "formatter_%s" % form
            opts = cp.options(sectname)
            if "format" in opts:
                fs = cp.get(sectname, "format", 1)
            else:
                fs = None
            if "datefmt" in opts:
                dfs = cp.get(sectname, "datefmt", 1)
            else:
                dfs = None
            f = logging.Formatter(fs, dfs)
            formatters[form] = f
    # next, do the handlers...
    # critical section...
    _acquireLock()
    try:
        try:
            # first, lose the existing handlers...
            _handlers.clear()
            # now set up the new ones...
            hlist = cp.get("handlers", "keys")
            if len(hlist):
                hlist = hlist.split(",")
                handlers = {}
                fixups = []  # for inter-handler references
                for hand in hlist:
                    sectname = "handler_%s" % hand
                    klass = cp.get(sectname, "class")
                    opts = cp.options(sectname)
                    if "formatter" in opts:
                        fmt = cp.get(sectname, "formatter")
                    else:
                        fmt = ""
                    klass = eval(klass)
                    args = cp.get(sectname, "args")
                    args = eval(args)
                    h = klass(args)
                    if "level" in opts:
                        lvl = cp.get(sectname, "level")
                        h.setLevel(_levelNames[lvl])
                    if len(fmt):
                        h.setFormatter(formatters[fmt])
                    # temporary hack for FileHandler and MemoryHandler.
                    if klass == logging.FileHandler:
                        maxsize = 0
                        if "maxsize" in opts:
                            ms = cp.getint(sectname, "maxsize")
                            if ms > 0:
                                maxsize = ms
                        if maxsize:
                            backcount = 0
                            if "backcount" in opts:
                                bc = cp.getint(sectname, "backcount")
                                if bc > 0:
                                    backcount = bc
                            h.setRollover(maxsize, backcount)
                    elif klass == logging.MemoryHandler:
                        if "target" in opts:
                            target = cp.get(sectname, "target")
                        else:
                            target = ""
                        if len(
                            target
                        ):  # the target handler may not be loaded yet,
                            # so keep for later...
                            fixups.append((h, target))
                    handlers[hand] = h
                # now all handlers are loaded, fixup inter-handler references..
                for fixup in fixups:
                    h = fixup[0]
                    t = fixup[1]
                    h.setTarget(handlers[t])
            # at last, the loggers...first the root...
            llist = cp.get("loggers", "keys")
            llist = llist.split(",")
            llist.remove("root")
            sectname = "printer_root"
            log = print_root
            opts = cp.options(sectname)
            if "level" in opts:
                lvl = cp.get(sectname, "level")
                log.setLevel(_levelNames[lvl])
            for h in print_root.handlers:
                print_root.removeHandler(h)
            hlist = cp.get(sectname, "handlers")
            if len(hlist):
                hlist = hlist.split(",")
                for hand in hlist:
                    log.addHandler(handlers[hand])
            # and now the others...
            # we don't want to lose the existing loggers,
            # since other threads may have pointers to them.
            # existing is set to contain all existing loggers,
            # and as we go through the new configuration we
            # remove any which are configured. At the end,
            # what's left in existing is the set of loggers
            # which were in the previous configuration but
            # which are not in the new configuration.

            existing = print_root.manager.printDict.keys()

            # now set up the new ones...
            for log in llist:
                sectname = "printer_%s" % log
                qn = cp.get(sectname, "qualname")
                opts = cp.options(sectname)
                if "propagate" in opts:
                    propagate = cp.getint(sectname, "propagate")
                else:
                    propagate = 1
                printer = getPrinter(qn)
                if qn in existing:
                    existing.remove(qn)
                if "level" in opts:
                    lvl = cp.get(sectname, "level")
                    printer.setLevel(_levelNames[lvl])
                for h in printer.handlers:
                    printer.removeHandler(h)
                printer.propagate = propagate
                printer.disabled = 0
                hlist = cp.get(sectname, "handlers")
                if len(hlist):
                    hlist = hlist.split(",")
                    for hand in hlist:
                        printer.addHandler(handlers[hand])
            # Disable any old loggers. There's no point deleting
            # them as other threads may continue to hold references
            # and by disabling them, you stop them doing any printing.

            for log in existing:
                print_root.manager.printerDict[log].disabled = 1
        except Exception:  # noqa: E722
            import traceback
            ei = sys.exc_info()
            traceback.print_exception(ei[0], ei[1], ei[2], None, sys.stderr)
            del ei
    finally:
        _releaseLock()


# ---------------------------------------------------------------------------
# Utility functions at module level.
# Basically delegate everything to the root printer.
# ---------------------------------------------------------------------------


def getPrinter(name=None):
    """
    Return a printer with the specified name, creating it if necessary.
    If no name is specified, return the root printer.
    """
    if name:
        return Printer.manager.getPrinter(name)
    else:
        return print_root


def job(msg, *args, **kwargs):
    """
    Log a message with severity 'JOB' on the root printer.
    """
    if len(print_root.handlers) == 0:
        basicConfig()
    print_root.job(msg, args, kwargs)


def important(msg, *args, **kwargs):
    """
    Log a message with severity 'IMPORTANT' on the root printer.
    """
    if len(print_root.handlers) == 0:
        basicConfig()
    print_root.important(msg, args, kwargs)


def terse(msg, *args, **kwargs):
    """
    Log a message with severity 'TERSE' on the root printer.
    """
    if len(print_root.handlers) == 0:
        basicConfig()
    print_root.terse(msg, args, kwargs)


def normal(msg, *args, **kwargs):
    """
    Log a message with severity 'NORMAL' on the root printer.
    """
    if len(print_root.handlers) == 0:
        basicConfig()
    print_root.normal(msg, args, kwargs)


def verbose(msg, *args, **kwargs):
    """
    Log a message with severity 'VERBOSE' on the root printer.
    """
    if len(print_root.handlers) == 0:
        basicConfig()
    print_root.verbose(msg, args, kwargs)


def disable(level):
    """
    Disable all printing calls less severe than 'level'.
    """
    print_root.manager.disable = level


def shutdown():
    """
    Perform any cleanup actions in the printing system (e.g. flushing
    buffers). Should be called at application exit.
    """
    for h in _handlers.keys():
        h.flush()
        h.close()


#
# Did not bring the socket listener code over yet
#
#


class FormattedText(object):
    """
    Class for wrapping, formatting and indenting text to make textual output
    from the MolSSI Framework and its plugins easier

    The __str__ method returns the formatted text, so e.g. print(<object>)
    formats the text and prints it as expected.

    Example:
    -----------------------------------------------------------------------
        ft = FormattedText('''\
        This is the first line.
        And the second.
        And a very, very, very, very, very, very, very, very, very, very, very, very long line.
        And i = '{i}' and j = '{j}' followed by a line break

        Ending with the last line. Which can of course also be wrapped
        if we want it to be.

            What about a list:     1

            With a second item:    2

        Does that work?''', i=1, j=2, indent='--->', line_length=70)

        print(ft)
    -----------------------------------------------------------------------
    Will print the following:

        --->This is the first line. And the second. And a very, very, very,
        --->very, very, very, very, very, very, very, very, very long line.
        --->And i = '1' and j = '2' followed by a line break

        --->Ending with the last line. Which can of course also be wrapped if
        --->we want it to be.

        --->    What about a list:     1

        --->    With a second item:    2

        --->Does that work?

    where the start of each arrow is aligned to the left (not indented as in
    this string).
    """  # noqa: E501

    def __init__(
        self,
        text,
        dedent=True,
        indent='',
        indent_initial=True,
        indent_all=False,
        wrap=True,
        line_length=80,
        *args,
        **kwargs
    ):
        """
        Handle the text <text>, which usually from either a simple string
        or a triple-quoted string.

        The <dedent> option uses textwrap.dedent to remove consistent initial
        spaces so that the string can be nicely lined up in the source code.

        <indent> provides an optional string to prefix each line of output

        <indent_all> indicates that empty line of those containing only white
        space also be indented. By the default the are not.

        <indent_initial> indicates that the first line shoud be indented.
        defaults to True.

        <wrap> asks that lines of text not starting with a blank be wrapped
        at <line_length>.

        <*args> and <**kwargs> are additional arguments for {}-type formatting
        in the text. This is done after dedenting but before wrapping the text.

        Empty lines separate 'paragraphs'. Each paragraph is wrapped by itself,
        which is the behavior that most people expect.

        Since lines starting with blanks (after dedent'ing) are not wrapped,
        this can be used for e.g. tables and lists.
        """

        self.text = text
        self.args = args
        self.kwargs = kwargs
        self.dedent = dedent
        self.indent = indent
        self.indent_initial = indent_initial
        self.indent_all = indent_all
        self.wrap = wrap
        self.line_length = line_length

    def __str__(self):
        """Turn into a formatted, wrapped string as requested.

        NB. If there are no extra args or kwargs, don't format the
        string. This avoids problems with strings that have, e.g.
        braces internally.
        """
        if self.dedent:
            if len(self.args) == 0 and len(self.kwargs) == 0:
                text = textwrap.dedent(self.text)
            else:
                text = textwrap.dedent(
                    self.text.format(*self.args, **self.kwargs)
                )
        else:
            if len(self.args) == 0 and len(self.kwargs) == 0:
                text = self.text
            else:
                text = self.text.format(*self.args, **self.kwargs)

        if self.wrap:
            wrapped_text = ''
            buffer = []
            if not self.indent_initial:
                initial_indent = ''
            else:
                initial_indent = self.indent
            for line in text.splitlines():
                if line.strip() == '' or line[0] == ' ':
                    if len(buffer) > 0:
                        wrapped_text += textwrap.fill(
                            '\n'.join(buffer),
                            self.line_length,
                            initial_indent=initial_indent,
                            subsequent_indent=self.indent
                        )
                        wrapped_text += '\n'
                        buffer = []
                    if line.strip() == '':
                        if self.indent_all:
                            wrapped_text += self.indent + '\n'
                        else:
                            wrapped_text += '\n'
                elif line.strip() != '' and line[0] != ' ':
                    buffer.append(line)
                if line.strip() != '' and line[0] == ' ':
                    wrapped_text += self.indent + line + '\n'
            if len(buffer) > 0:
                wrapped_text += textwrap.fill(
                    '\n'.join(buffer),
                    self.line_length,
                    initial_indent=initial_indent,
                    subsequent_indent=self.indent
                )

            return wrapped_text
        else:
            if self.indent_all:
                return textwrap.indent(text, self.indent, lambda line: True)
            else:
                return textwrap.indent(text, self.indent)


if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package."""

import seamm_util
import os
import os.path


def test_open_1file():
    """Testing that we can read a simple file
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file1.txt.bz2')
    data = ['file1 line 1', 'file1 line 2', 'file1 line 3', 'file1 line 4']

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_open_2files():
    """Testing that the default 'include' works
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file_middle.txt.bz2')
    data = [
        'file_middle line 1', 'file_middle line 2', 'file1 line 1',
        'file1 line 2', 'file1 line 3', 'file1 line 4', 'file_middle line 3'
    ]

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_open_2files_at_end():
    """Testing that the default 'include' works at the end of a file
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file_end.txt.bz2')
    data = [
        'file_end line 1', 'file_end line 2', 'file_end line 3',
        'file1 line 1', 'file1 line 2', 'file1 line 3', 'file1 line 4'
    ]

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_lines():
    """Testing the line count for each line, with an included file

    Note that the line count includes the line with 'include'
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file_middle.txt.bz2')

    num = [1, 2, 1, 2, 3, 4, 4]
    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert fd.lineno == num[i]
            i += 1


def test_total_lines():
    """Testing the total number of lines with an included file

    Note that the line count includes the line with 'include'
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file_middle.txt.bz2')

    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            pass
        assert fd.total_lines == 8


def test_nested_includes():
    """Testing an include within an include
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file_include1.txt.bz2')
    data = [
        'file_include1 line1', 'file_end line 1', 'file_end line 2',
        'file_end line 3', 'file1 line 1', 'file1 line 2', 'file1 line 3',
        'file1 line 4'
    ]

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_include_empty_file():
    """Testing that including an empty file works
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'include_empty_file.txt.bz2')
    data = ['file_end line 1', 'file_end line 2', 'file_end line 3']

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_blank_lines():
    """Testing that empty lines and ones with just blanks don't
    cause any problems
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'include_blank_lines.txt.bz2')
    data = [
        'file_end line 1', '', 'file_end line 3', 'blank_lines first line',
        '   ', '', 'blank_lines 4th line', '   ', ''
    ]

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert line[0:-1] == data[i]
            i += 1


def test_stack():
    """Testing the stack of files part way through reading includes
    """
    cwd = os.path.dirname(__file__)
    datapath = os.path.join(cwd, 'data')
    filepath = os.path.join(datapath, 'file_include1.txt.bz2')
    data = [
        os.path.normcase('data/file1.txt.bz2:2'),
        os.path.normcase('data/file_end.txt.bz2:4')
    ]

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            i += 1
            if i == 6:
                stack = []
                for line in fd.stack():
                    stack.append(os.path.relpath(line, cwd))
                assert stack == data
                break


def test_different_include():
    """Testing that '#include' works
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file_#include.txt.bz2')
    data = [
        'file_middle line 1', 'file_middle line 2', 'file1 line 1',
        'file1 line 2', 'file1 line 3', 'file1 line 4', 'file_middle line 3'
    ]

    i = 0
    with seamm_util.Open(filepath, 'r', include='#include') as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_push():
    """Testing that the default 'include' works
    """
    datapath = os.path.join(os.path.dirname(__file__), 'data')
    filepath = os.path.join(datapath, 'file_middle.txt.bz2')
    data = [
        'file_middle line 1', 'file_middle line 2', 'file1 line 1',
        'file1 line 2', 'file1 line 3', 'file1 line 4', 'file_middle line 3',
        'file1 line 4', 'file_middle line 3'
    ]

    i = 0
    with seamm_util.Open(filepath, 'r', include='include') as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1
            if i == 7:
                fd.push(2)



# -*- coding: utf-8 -*-

"""Helpful routine using the underlying Python functions
in 'statistics' and 'statsmodels' to handle common needs
in molecular modeling
"""

import logging
import random
import statistics
import statsmodels.tsa.stattools as stattools

logger = logging.getLogger(__name__)


def analyze_autocorrelation(
    y, interval=1, nlags=64, method='zr', use_confidence=False
):
    """Find the statistical inefficiency, correlation time and other useful
    parameters given the time sequence of values 'y'.

    This function implements two approaches, 'zr' and 'cspsd' corresponding to
    the two papers (using the initials as the method name):

    zr: Zięba, A.; Ramza, P. Standard Deviation of the Mean of Autocorrelated
    Observations Estimated with the Use of the Autocorrelation Function
    Estimated From the Data. Metrology and Measurement Systems 2011, 18 (4),
    3–14.

    cspsd: Chodera, J. D.; Swope, W. C.; Pitera, J. W.; Seok, C.; Dill,
    K. A. Use of the Weighted Histogram Analysis Method for the Analysis of
    Simulated and Parallel Tempering Simulations. J. Chem. Theory Comput. 2006,
    3 (1), 26–41.

    Args:
        y ([float]): the time sequence to analyze
        interval (float): the time interval between values, defaults to 1
        nlags (int): the number of lags to start with
        method (str): the approach to use, 'zr' or 'cspsd'. Defaults to 'zr'

    Returns:
        dict(
            'n': (int) number of values in y
            'n_effective': (int) effective number of uncorrelated values
            'n_c': (int) the first zero crossing in the ACF
            'n_tau': (int) number of intervals in the correlation time
            'tau': (float) the correlation time if <interval> is correct
            'inefficiency': (float) the statistical inefficiency
            'acf': ([float]) the autocorrelation function
            'confidence_interval': [(float, float)] the 95% confidence interval

    The input parameter <nlags> is solely for performance. Since calculating
    the ACF is expensive (O(N^2)) and often the correlation time is short, at
    least if the trajectory is sampled reasonably, this function starts with a
    small number of lags, and doubles the number repeatedly if it does not find
    the zero crossing it expects.
    """

    # Find the autocorrelation time...
    n = len(y)

    logger.debug('analyze_autocorrelation for a vector of length {}'.format(n))

    if nlags >= n:
        nlags = n - 1

    while True:
        logger.debug('   nlags = {}'.format(nlags))

        acf, confidence = stattools.acf(
            y, nlags=nlags, alpha=0.05, fft=nlags > 16, unbiased=False
        )

        # remove the first items, which are 1 by definition
        acf = acf[1:]
        confidence = confidence[1:]

        # Find the last lag that is > 0
        sum_acf = 0.0
        n_c = 0
        if use_confidence:
            for lower, upper in confidence:
                delta = acf[n_c] - lower
                value = lower + delta * 3 / 4
                if nlags == n - 1:
                    print(value)
                if value < 0:
                    break
                sum_acf += acf[n_c]
                n_c += 1
        else:
            for value in acf:
                if nlags == n - 1:
                    print(value)
                if value < 0:
                    break
                sum_acf += value
                n_c += 1

        # If we found n_c, go on. Otherwise double the lags
        if value < 0:
            break
        else:
            if nlags == n - 1:
                raise RuntimeError(
                    'analyze_autocorrelation: Serious error! '
                    'Did not find negative autocorrelation value.'
                )
            nlags *= 2
            if nlags >= n:
                nlags = n - 1

    logger.debug('   n_c = {}'.format(n_c))

    if method == 'zr':
        # Use the approach of Zięba and Ramza
        n_eff = (n - 2 * n_c - 1 + n_c * (n_c + 1) / n) / (1 + 2 * sum_acf)
        inefficiency = n / n_eff
        n_tau = (inefficiency - 1) / 2
    else:
        # Use the approach of Chodera, Swope, Pitera, Seok and Dill
        n_tau = 0.0
        for t in range(0, n_c):
            n_tau += (1 - (t + 1) / n) * acf[t]

        inefficiency = 1 + 2 * n_tau
        n_eff = n / inefficiency

    tau = n_tau * interval

    result = {
        'n': n,
        'n_effective': n_eff,
        'n_c': n_c,
        'n_tau': n_tau,
        'tau': tau,
        'inefficiency': inefficiency,
        'acf': acf,
        'confidence_interval': confidence
    }

    return result


def ar1(n=1000, a=10.0, b=0.2, sigma=0.5, seed=None):
    """Generate an AR(1) series of length n
    """
    r = random.Random()
    r.seed(seed)

    y0 = a / (1 - b)
    y = [0.0] * n
    y[0] = a + b * y0 + sigma * (r.random() - 0.5)
    for i in range(1, n):
        y[i] = a + b * y[i - 1] + sigma * (r.random() - 0.5)

    return y


if __name__ == "__main__":
    print('in end section')
    import time

    print()

    a = 10.0
    b = 0.9
    for n in (15, 60, 240, 512, 1024, 2048, 10000, 100000):
        y = ar1(n, a=a, b=b, seed=52)
        t0 = time.time()
        r1 = analyze_autocorrelation(y, method='zr', nlags=4)
        t1 = time.time()
        r2 = analyze_autocorrelation(
            y, method='zr', nlags=4, use_confidence=True
        )
        t2 = time.time()

        print('{:>20s} = {:7.3f} {:7.3f}'.format('time', t1 - t0, t2 - t1))
        print(
            '{:>20s} = {} {}'.format('nlags', len(r1['acf']), len(r2['acf']))
        )
        for key in r1:
            if key not in ['acf', 'confidence_interval']:
                print(
                    '{:>20s} = {:7.1f} {:7.1f}'.format(key, r1[key], r2[key])
                )

        ave = statistics.mean(y)
        stdev = statistics.stdev(y)
        print('                mean = {:9.3f}'.format(a / (1 - b)))
        print('                 ave = {:9.3f}'.format(ave))
        print('               stdev = {:9.3f}'.format(stdev))
        print()



# -*- coding: utf-8 -*-

"""Functions for handling executables"""

import logging
import os
import os.path

logger = logging.getLogger(__name__)


def check_executable(executable):
    """Check that an executable exists.

    Check that an executable exists and is executable. If no path is
    given, but just the executable name, then the current path is
    checked.  If the executable is missing or not actually executable
    this function raises a useful error with a message letting the
    user know where the path came from -- a command line option, an
    environment variable, configuration file, or default.

    Parameters:
        executable -- required positional parameter giving the name of
            the executable as a string. It can be a simple name or contain
            a path to the executable.

    Returns:
        The string for the path to the executable. If it was found on the path,
        the full path is returned, otherwise the return is the same as the
        executable string passed in.

    Errors:
        FileNotFoundError if the executable is not found or is not executable.
            The message contains information about where the option came from,
            either the command-line, an environment variable, configuration
            file or the default.
    """

    if os.path.dirname(executable) == '':
        # A simple name without a directory part was given, so see if the
        # executable exists in the path.
        path = os.get_exec_path()
        exes = []
        for dir in path:
            exe = os.path.join(dir, executable)
            if os.path.isfile(exe):
                if os.access(exe, os.X_OK):
                    return exe
                else:
                    exes.append(exe)
        msg = (
            "The executable '{}' was not found in the path:\n\t"
            .format(executable)
        )
        msg += '\n\t'.join(path)
        msg += '\n'
        if len(exes) > 0:
            msg += (
                'However, the following files were found, but were not '
                'executable:\n\t'
            )
            msg += '\n\t'.join(exes)
            msg += '\n'
    else:
        #  A path before the executable name was given, so try as is
        if os.path.isfile(executable):
            if os.access(executable, os.X_OK):
                return executable
            else:
                msg = (
                    "The executable '{}' exists, but is not executable!\n"
                    .format(executable)
                )
        else:
            msg = "The executable '{}' does not exist!\n".format(executable)

    raise FileNotFoundError(msg)
